'''
segy header item
'''

# import python standard modules


# import 3rd party libraries


# import local python


class SegyHeaderItem(object):
    """Used to describe a single header item or field in a segy file header or trace header.

    Parameters
    ----------
    name : str
        Header item identifier.
    nbytes : int
        number of bytes.
    startbyte : int
        Zero-indexed starting location.
    description : str
        Description of the header item.
    signed : bool
        If the header item is an int, this flag indicates that the bytes object is signed.
    map_bool : bool
        Flag that indicates where this item is to be mapped from the interpreted value to another value.
    map_dict : `dict`
        Dictionary in which the key is the interpreted value. Required if map_bool is True.
    value : Numeric
        Header item value.
    """

    def __init__(self, name:str, nbytes=4, startbyte=0, description='', signed=False,
                 map_bool=False, map_dict={}, value=0):
        self.name = name
        self.nbytes = nbytes            # length of the header in bytes; default to 4, but may be 2
        self.startbyte = startbyte      # the relative byte location to start reading
        self.description = description  # defualt to empty string
        self.signed = signed            # a boolean; True is a signed twos complement integer, False is not signed
        self.value = value              # the value obtained from the segy file
        self.supplementary = ''         # any additional supplementary material
        self.map_bool = map_bool        # True if value is mapped to a dict; False otherwise
        self.map_dict = map_dict        # name of the dictionary to map to

    def __str__(self):
        s = 'name: ' + self.name + ', '
        s += 'description: ' + self.description + ', '
        s += 'start byte: ' + str(self.startbyte) + ', '
        s += 'byte length: ' + str(self.nbytes) + ', '
        s += 'value: ' + str(self.value)
        return s

    # ACCESSORS
    def field(self, field_id):
        if field_id == 'name':
            return self.name
        elif field_id == 'nbytes':
            return self.nbytes
        elif field_id == 'startbyte':
            return self.startbyte
        elif field_id == 'description':
            return self.description
        elif field_id == 'signed':
            return self.signed
        elif field_id == 'value':
            return self.value
        elif field_id == 'supplementary':
            return self.supplementary
        elif field_id == 'map_bool':
            return self.map_bool
        elif field_id == 'map_dict':
            return self.map_dict
        else:
            print('field not found')
            return None

    def header_name(self):
        return self.name

    def number_of_bytes(self):
        return self.nbytes

    def start_byte(self):
        return self.startbyte

    # def description(self):
    #     return self.description

    def is_signed(self):
        return self.signed

    # def value(self):
    #     return self.value

    # def supplementary(self):
    #     return self.supplementary

    # MUTATORS 
    def set_field(self, field, value):
        if field == 'name':
            self.name = value
        elif field == 'nbytes':
            self.nbytes = value
        elif field == 'startbyte':
            self.startbyte = value
        elif field == 'description':
            self.description = value
        elif field == 'signed':
            self.signed = value
        elif field == 'value':
            self.value = value
        elif field == 'supplementary':
            self.supplementary = value
        elif field == 'map_bool':
            self.map_bool = value
        elif field == 'map_dict':
            self.map_dict = value
        else:
            print('field not found')


    def set_name(self, name:str):
        self.name = name

    def set_number_of_bytes(self, nbytes:int):
        self.nbytes = nbytes

    def set_start_byte(self, startbyte:int):
        self.startbyte = startbyte

    def set_description(self, description:str):
        self.description = description

    def set_signed(self, signed:bool):
        self.signed = signed

    def set_value(self, value):
        self.value = value

    def set_supplementary_text(self, supplementary:str):
        self.supplementary = supplementary


"""
SPS parser
"""
from pandas import DataFrame

# H26
# H26      Relation Record Specification
# H26
# H26 DEFINITION OF FIELD            COLS       FORMAT       UNIT
# H26 Record identification          1- 1           A1       none
# H26 Line name (l adj)              2-17          4A4       none
# H26 Point number (right adj)      18-25          2A4       none
# H26 Point index                   26-26           I1       none
# H26 Point code                    27-28           A2       none
# H26 Static correction             29-32           I4       Msec
# H26 Point depth                   33-36          F4.1      Metre
# H26 Seismic datum                 37-40           I4       Metre
# H26 Uphole time                   41-42           I2       Msec
# H26 Water depth                   43-46          F4.1      Metre
# H26 Map grid easting              47-55          F9.1      none
# H26 Map grid northing             56-65          F10.1     none
# H26 Surface elevation             66-71          F6.1      Metre
# H26 Day of year                   72-74           I3       none
# H26 Time hhmmss                   75-80          3I2       none
# H26

class SPSLineItem(object):
    def __init__(self, s: str):
        self.line_name = s[1:17].strip()
        self.point_number = s[17:25].strip()
        self.point_index = s[25:26].strip()
        self.point_code = s[26:28].strip()
        self.static_correction = s[28:32].strip()
        self.point_depth = s[32:36].strip()
        self.seismic_datum = s[36:40].strip()
        self.uphole_time = s[40:42].strip()
        self.water_depth = s[42:46].strip()
        self.map_grid_easting = s[46:55].strip()
        self.map_grid_northing = s[55:65].strip()
        self.surface_elevation = s[65:71].strip()

    def __str__(self):
        srep = '\nsource point\n'
        srep += "Line name         : " + self.line_name + '\n'
        srep += "Point number      : " + self.point_number + '\n'
        srep += "Point index       : " + self.point_index + '\n'
        srep += "Point code        : " + self.point_code + '\n'
        srep += "Static correction : " + self.static_correction + '\n'
        srep += "Point depth       : " + self.point_depth + '\n'
        srep += "Seismic datum     : " + self.seismic_datum + '\n'
        srep += "Uphole time       : " + self.uphole_time + '\n'
        srep += "Water depth       : " + self.water_depth + '\n'
        srep += "Map grid easting  : " + self.map_grid_easting + '\n'
        srep += "Map grid northing : " + self.map_grid_northing + '\n'
        srep += "Surface elevation : " + self.surface_elevation
        return srep

class SPS(object):
    def __init__(self, f: str):
        line_names = []
        point_numbers = []
        point_indices = []
        point_codes = []
        static_corrections = []
        point_depths = []
        seismic_datums = []
        uphole_times = []
        water_depths = []
        map_grid_eastings = []
        map_grid_northings = []
        surface_elevations = []
        with open(f, "r") as fobj:
            for cnt, line in enumerate(fobj):
                if not line.startswith("H26"):
                    record = SPSLineItem(line)
                    line_names.append(record.line_name)
                    point_numbers.append(int(record.point_number))
                    point_indices.append(int(record.point_index))
                    point_codes.append(record.point_code)
                    static_corrections.append(float(record.static_correction))
                    point_depths.append(float(record.point_depth))
                    seismic_datums.append(float(record.seismic_datum))
                    uphole_times.append(float(record.uphole_time))
                    water_depths.append(float(record.water_depth))
                    map_grid_eastings.append(float(record.map_grid_easting))
                    map_grid_northings.append(float(record.map_grid_northing))
                    surface_elevations.append(float(record.surface_elevation))

            fobj.close()
        # temporary dict for dataframe constructor
        tmp_dict = dict()
        tmp_dict['line_name'] = line_names
        tmp_dict['point_number'] = point_numbers
        tmp_dict['point_index'] = point_indices
        tmp_dict['point_code'] = point_codes
        tmp_dict['static_correction'] = static_corrections
        tmp_dict['point_depth'] = point_depths
        tmp_dict['seismic_datum'] = seismic_datums
        tmp_dict['uphole_time'] = uphole_times
        tmp_dict['water_depth'] = water_depths
        tmp_dict['map_grid_easting'] = map_grid_eastings
        tmp_dict['map_grid_northing'] = map_grid_northings
        tmp_dict['surface_elevation'] = surface_elevations
        self._df = DataFrame(tmp_dict)

    def dataframe(self):
        return self._df

    def source_x(self, source_point: int):
        tmpi = self._df.loc[self._df['point_number'] == source_point].index.values
        return self._df.loc[tmpi[0], 'map_grid_easting']

    def source_y(self, source_point: int):
        tmpi = self._df.loc[self._df['point_number'] == source_point].index.values
        return self._df.loc[tmpi[0], 'map_grid_northing']

    def source_elevation(self, source_point: int):
        tmpi = self._df.loc[self._df['point_number'] == source_point].index.values
        return self._df.loc[tmpi[0], 'surface_elevation']

    def source_static(self, source_point: int):
        tmpi = self._df.loc[self._df['point_number'] == source_point].index.values
        return self._df.loc[tmpi[0], 'static_correction']

# END


"""Mappings between the coding systems used for sample types.
"""

# A mapping from data sample format codes to SEG Y types.
from collections import namedtuple
from enum import IntEnum, Enum

from segytools.ibm_float import MIN_IBM_FLOAT, MAX_IBM_FLOAT


class DataSampleFormat(IntEnum):
    """Data sample format code. Mandatory for all data.
    1 = 4-byte IBM floating-point,
    2 = 4-byte, two's complement integer,
    3 = 2-byte, two's complement integer,
    4 = 4-byte fixed-point with gain (obsolete),
    5 = 4-byte IEEE floating-point,
    6 = Not currently used,
    7 = Not currently used,
    8 = 1-byte, two's complement integer."""
    IBM = 1
    INT32 = 2
    INT16 = 3
    FLOAT32 = 5
    INT8 = 8


class SegYType(str, Enum):
    INT32 = 'int32'
    NNINT32 = 'nnint32'
    INT16 = 'int16'
    NNINT16 = 'nnint16'
    INT8 = 'int8'
    NNINT8 = 'nnint8'
    FLOAT32 = 'float32'
    IBM = 'ibm'


DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE = {
    DataSampleFormat.IBM: SegYType.IBM,
    DataSampleFormat.INT32: SegYType.INT32,
    DataSampleFormat.INT16: SegYType.INT16,
    DataSampleFormat.FLOAT32: SegYType.FLOAT32,
    DataSampleFormat.INT8: SegYType.INT8}


SEG_Y_TYPE_TO_DATA_SAMPLE_FORMAT = {v: k for k, v in DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE.items()}


# A mapping from SEG Y data types to format characters used by the
# Python Standard Library struct module
SEG_Y_TYPE_TO_CTYPE = {
    SegYType.INT32:  'i',
    SegYType.NNINT32: 'I',
    SegYType.INT16:  'h',
    SegYType.NNINT16: 'H',
    SegYType.INT8: 'b',
    SegYType.NNINT8: 'B',
    SegYType.FLOAT32:  'f',
    SegYType.IBM: 'ibm'}


# Human readable descriptions of the sample types.
SEG_Y_TYPE_DESCRIPTION = {
    SegYType.IBM: 'IBM 32 bit float',
    SegYType.INT32: '32 bit signed integer',
    SegYType.NNINT32: '32 bit non-negative integer',
    SegYType.INT16: '16 bit signed integer',
    SegYType.NNINT16: '16 bit non-negative integer',
    SegYType.FLOAT32: 'IEEE float32',
    SegYType.INT8: '8 bit signed integer (byte)',
    SegYType.NNINT8: '8 bit unsigned integer (byte)'}


# Sizes of various ctypes in bytes
CTYPE_TO_SIZE = dict(
    i=4,
    I=4,
    h=2,
    H=2,
    b=1,
    B=1,
    f=4,
    ibm=4)


def data_sample_format_size_in_bytes(data_sample_format: DataSampleFormat):
    """The size in bytes per sample of a data sample format.
    """
    return CTYPE_TO_SIZE[SEG_Y_TYPE_TO_CTYPE[DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE[data_sample_format]]]


def data_sample_format_description(data_sample_format: DataSampleFormat):
    """A textual description of the data sample format.
    """
    return SEG_Y_TYPE_DESCRIPTION[DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE[data_sample_format]]


def size_in_bytes(ctype):
    """The size in bytes of a ctype.
    """
    try:
        return CTYPE_TO_SIZE[ctype]
    except KeyError:
        raise ValueError("No such C-type {!r}".format(ctype))

Limits = namedtuple('Limits', ['min', 'max'])

LIMITS = {
    SegYType.IBM: Limits(MIN_IBM_FLOAT, MAX_IBM_FLOAT),
    SegYType.INT32: Limits(-2147483648, 2147483647),
    SegYType.NNINT32: Limits(0, 2147483647),
    SegYType.INT16: Limits(-32768, 32767),
    SegYType.NNINT16: Limits(0, 32767),
    SegYType.FLOAT32: Limits(-3.402823e38, 3.402823e38),
    SegYType.INT8: Limits(-128, 127),
    SegYType.NNINT8: Limits(0, 127)
}

PY_TYPES = {
    SegYType.IBM: float,
    SegYType.INT32: int,
    SegYType.NNINT32: int,
    SegYType.INT16: int,
    SegYType.NNINT16: int,
    SegYType.FLOAT32: float,
    SegYType.INT8: int,
    SegYType.NNINT8: int
}

ENDIAN = {
    '<': 'little',
    '>': 'big'
}


"""Collection of tools for working with segy data"""

# standard
from array import array

# third party

# local
from segytools.ibm_float import IBMFloat

def unpack_ibm_floats(data, num_items):
    """Unpack a series of binary-encoded big-endian single-precision IBM floats.

    Args:
        data: A sequence of bytes.

        num_items: The number of floats to be read.

    Returns:
        A sequence of floats.
    """
    return [IBMFloat.from_bytes(data[i: i+4]) for i in range(0, num_items * 4, 4)]


def unpack_values(buf, ctype, endian='big'):
    """Unpack a series items from a byte string.

    Args:
        data: A sequence of bytes.

        ctype: A format code (one of the values in the datatype.CTYPES
            dictionary)

        endian: '>' for big-endian data (the standard and default), '<'
            for little-endian (non-standard)

    Returns:
        A sequence of objects with type corresponding to the format code.
    """
    a = array(ctype, buf)
    if endian != sys.byteorder:
        a.byteswap()
    return a

def read_binary_values(buf, num_items, ctype, endian):
    """Read trace data from bytes object and convert to list of floats

    Parameters
    ----------
    buf : bytes
        buffer of byte data
    num_items : int
        number of samples expected in output
    ctype : str
        character type; see segpy.datatypes.SEG_Y_TYPE_TO_CTYPE for acceptable values
    endian : str
        either 'little' or 'big'
    """
    values = (unpack_ibm_floats(buf, num_items)
              if ctype == 'ibm'
              else unpack_values(buf, ctype, endian))
    assert len(values) == num_items
    return values


from math import frexp, isnan, isinf, trunc, floor, gcd
from numbers import Real, Rational

from segytools.utils import four_bytes


IBM_ZERO_BYTES = b'\x00\x00\x00\x00'
IBM_NEGATIVE_ONE_BYTES = b'\xc1\x10\x00\x00'
IBM_POSITIVE_ONE_BYTES = b'A\x10\x00\x00'

MIN_IBM_FLOAT = -7.2370051459731155e+75
LARGEST_NEGATIVE_NORMAL_IBM_FLOAT = -5.397605346934028e-79
SMALLEST_POSITIVE_NORMAL_IBM_FLOAT = 5.397605346934028e-79
MAX_IBM_FLOAT = 7.2370051459731155e+75

MAX_BITS_PRECISION_IBM_FLOAT = 24
MIN_BITS_PRECISION_IBM_FLOAT = 21  # The first 3 bits of the mantissa may be zero
EPSILON_IBM_FLOAT = pow(2.0, -(MIN_BITS_PRECISION_IBM_FLOAT - 1))
_L24 = 2 ** MAX_BITS_PRECISION_IBM_FLOAT
_F24 = float(pow(2, MAX_BITS_PRECISION_IBM_FLOAT))

_L21 = 2 ** MIN_BITS_PRECISION_IBM_FLOAT

EXPONENT_BASE = 16
EXPONENT_BIAS = 64

MIN_EXACT_INTEGER_IBM_FLOAT = -2**MAX_BITS_PRECISION_IBM_FLOAT
MAX_EXACT_INTEGER_IBM_FLOAT = 2**MIN_BITS_PRECISION_IBM_FLOAT


def ibm2ieee(big_endian_bytes):
    """Interpret a byte string as a big-endian IBM float.

    Args:
        big_endian_bytes (bytes): A byte-string containing at least four bytes.

    Returns:
        The floating point value.
    """
    a, b, c, d = four_bytes(big_endian_bytes)

    if b == c == d == 0:
        return 0.0

    sign = -1 if (a & 0x80) else 1
    exponent_16_biased = a & 0x7f
    mantissa = ((b << 16) | (c << 8) | d) / _F24

    value = sign * mantissa * pow(EXPONENT_BASE, exponent_16_biased - EXPONENT_BIAS)
    return value


BITS_PER_NYBBLE = 4


def ieee2ibm(f):
    """Convert a float to four big-endian bytes representing an IBM float.

    Args:
        f (float): The value to be converted.

    Returns:
        A bytes object (Python 3) or a string (Python 2) containing four
        bytes representing a big-endian IBM float.

    Raises:
        OverflowError: If f is outside the representable range.
        ValueError: If f is NaN or infinite.
        FloatingPointError: If f cannot be represented without total loss of precision.
    """
    if f == 0:
        # There are many potential representations of zero - this is the standard one
        return b'\x00\x00\x00\x00'

    if isnan(f):
        raise ValueError("NaN cannot be represented in IBM floating point")

    if isinf(f):
        raise ValueError("Infinities cannot be represented in IBM floating point")

    if f < MIN_IBM_FLOAT:
        raise OverflowError("IEEE Floating point value {} is less than the "
                            "representable minimum for IBM floats.".format(f))

    if f > MAX_IBM_FLOAT:
        raise OverflowError("IEEE Floating point value {} is greater than the "
                            "representable maximum for IBM floats".format(f))

    # Now compute m and e to satisfy:
    #
    #  f = m * 2^e
    #
    # where 0.5 <= abs(m) < 1
    # except when f == 0 in which case m == 0 and e == 0, which we've already
    # dealt with.
    m, e = frexp(f)

    # Convert the fraction (m) into an integer representation. IEEE float32
    # numbers have 23 explicit (24 implicit) bits of precision.

    mantissa = abs(int(m * _L24))
    exponent = e
    sign = 0x80 if f < 0 else 0x00

    # IBM single precision floats are of the form
    # (-1)^sign * 0.significand * 16^(exponent-64)

    # Adjust the exponent, and the mantissa in sympathy so it is
    # a multiple of four, so it can be expressed in base 16
    remainder = exponent % 4
    if remainder != 0:
        shift = 4 - remainder
        mantissa >>= shift
        exponent += shift

    exponent_16 = exponent >> 2            # Divide by four to convert to base 16
    exponent_16_biased = exponent_16 + EXPONENT_BIAS  # Add the exponent bias of 64

    # If the biased exponent is negative, we try to use a subnormal representation
    if exponent_16_biased < 0:
        shift_16 = 0 - exponent_16_biased
        exponent_16_biased += shift_16  # An increment of the base-16 exponent must be balanced by
        mantissa >>= 4 * shift_16       # A division by 16 (four binary places) in the mantissa
        if mantissa == 0:
            raise FloatingPointError("IEEE Floating point value {} is smaller than the "
                                     "smallest subnormal number for IBM floats.".format(f))

    a = sign | exponent_16_biased
    b = (mantissa >> 16) & 0xff
    c = (mantissa >> 8) & 0xff
    d = mantissa & 0xff

    return bytes((a, b, c, d))


class IBMFloat(Real):

    __slots__ = ['_data']

    def __new__(cls, b):
        obj = object.__new__(cls)

        data = bytes(b)
        num_bytes = len(data)
        if num_bytes != 4:
            raise ValueError("{} cannot be constructed from {} values".format(cls.__name__, num_bytes))
        obj._data = data
        return obj

    @classmethod
    def from_float(cls, f):
        """Construct an IBMFloat from an IEEE float.

        Args:
            f (float): The value to be converted.

        Returns:
            An IBMFloat.

        Raises:
            OverflowError: If f is outside the representable range.
            ValueError: If f is NaN or infinite.
            FloatingPointError: If f cannot be represented without total loss of precision.
        """
        if isinstance(f, IBMFloat):
            return f
        return cls(ieee2ibm(f))

    @classmethod
    def from_float_without_underflow(cls, f):
        """Construct an IBMFloat from an IEEE float.

        Args:
            f (float): The value to be converted. If the provided
                IEEE value underflows the smallest representable
                IBM value, this function returns zero.

        Returns:
            An IBMFloat.

        Raises:
            OverflowError: If f is outside the representable range.
            ValueError: If f is NaN or infinite.
        """
        try:
            buffer = ieee2ibm(f)
        except FloatingPointError:  # Underflow
            return IBM_FLOAT_ZERO
        else:
            return cls(buffer)

    @classmethod
    def from_real(cls, f):
        return cls.from_float(f)

    @classmethod
    def from_bytes(cls, b):
        return cls(b)

    @classmethod
    def ldexp(cls, fraction, exponent):
        """Make an IBMFloat from fraction and exponent.

        The is the inverse function of IBMFloat.frexp()

        Args:
            fraction: A Real in the range -1.0 to 1.0.
            exponent: An integer in the range -256 to 255 inclusive.
        """
        if not (-1.0 <= fraction <= 1.0):
            raise ValueError("ldexp fraction {!r} out of range -1.0 to +1.0")

        if not (-256 <= exponent < 256):
            raise ValueError("ldexp exponent {!r} out of range -256 to 256")

        ieee = fraction * 2**exponent
        return IBMFloat.from_float(ieee)

    @property
    def signbit(self):
        """True if the value is negative, otherwise False."""
        return bool(self._data[0] & 0x80)

    def __float__(self):
        return ibm2ieee(self._data)

    def __bytes__(self):
        return self._data

    def __repr__(self):
        return "{}(bytes([{}])) ≈ {!r}".format(
            self.__class__.__name__,
            ', '.join('0x{:02x}'.format(b) for b in self._data),
            float(self))

    def __str__(self):
        return str(float(self))

    def __bool__(self):
        return not self.is_zero()

    def is_zero(self):
        return self.int_mantissa == 0

    def is_subnormal(self):
        if self.is_zero():
            # Only one of the many possible representations of zero is considered 'normal' - all the zeros
            return not all(b == 0 for b in self._data)

        return self._data[1] < 16

    def zero_subnormal(self):
        return IBM_FLOAT_ZERO if self.is_subnormal() else self

    def frexp(self):
        """Obtain the fraction and exponent.

        Returns:
            A pair where the first item is the fraction in the range -1.0 and +1.0 and the
            exponent is an integer such that f = fraction * 2**exponent
        """
        sign = -1 if self.signbit else 1
        mantissa = sign * self.int_mantissa / _F24
        exp_2 = self.exp16 * 4
        return mantissa, exp_2

    def as_integer_ratio(self):
        sign = -1 if self.signbit else 1
        e16 = self.exp16
        if e16 >= 0:
            numerator = self.int_mantissa * EXPONENT_BASE**self.exp16
            denominator = _L24
        else:
            numerator = self.int_mantissa
            denominator = _L24 * EXPONENT_BASE**abs(self.exp16)
        divisor = gcd(numerator, denominator)
        reduced_numerator = sign * numerator // divisor
        reduced_denominator = denominator // divisor
        return reduced_numerator, reduced_denominator

    def __pos__(self):
        return self

    def __neg__(self):
        if self.is_zero():
            return IBM_FLOAT_ZERO

        data = self._data
        return IBMFloat((data[0] ^ 0b10000000,
                         data[1],
                         data[2],
                         data[3]))

    def __abs__(self):
        if self.is_zero():
            return IBM_FLOAT_ZERO

        data = self._data
        return IBMFloat((data[0] & 0b01111111,
                         data[1],
                         data[2],
                         data[3]))

    def __eq__(self, rhs):
        lhs = self

        if lhs is rhs:
            return True

        if isinstance(rhs, float):
            if isnan(rhs) or isinf(rhs):
                return 0.0 == rhs
            lhs_numerator, lhs_denominator = lhs.as_integer_ratio()
            rhs_numerator, rhs_denominator = rhs.as_integer_ratio()
            return (lhs_numerator == rhs_numerator) and (lhs_denominator == rhs_denominator)

        if isinstance(rhs, Rational):
            lhs_numerator, lhs_denominator = lhs.as_integer_ratio()
            return (lhs_numerator == rhs.numerator) and (lhs_denominator == rhs.denominator)

        if not isinstance(rhs, IBMFloat):
            return NotImplemented

        if lhs._data == rhs._data:
            return True

        lhs_sign = lhs.signbit
        rhs_sign = rhs.signbit

        if lhs_sign != rhs_sign:
            return False

        nlhs = lhs.try_normalize()
        nrhs = rhs.try_normalize()

        if not (nlhs.is_subnormal() or nrhs.is_subnormal()):
            # Both of the numbers are normalised
            return nlhs._data == nrhs._data

        # Either or both of the numbers are subnormal
        lhs_exp16 = nlhs.exp16
        rhs_exp16 = nrhs.exp16

        lhs_mantissa = nlhs.int_mantissa
        rhs_mantissa = nrhs.int_mantissa

        if lhs_exp16 < rhs_exp16:
            delta_exp16 = rhs_exp16 - lhs_exp16
            lhs_mantissa >>= 4 * delta_exp16
            lhs_exp16 += delta_exp16

        if lhs_exp16 > rhs_exp16:
            delta_exp16 = lhs_exp16 - rhs_exp16
            rhs_mantissa >>= 4 * delta_exp16
            rhs_exp16 += delta_exp16

        assert lhs_exp16 == rhs_exp16
        return lhs_mantissa == rhs_mantissa

    @property
    def exp16(self):
        """The base 16 exponent."""
        exponent_16_biased = self._data[0] & 0x7f
        exponent_16 = exponent_16_biased - EXPONENT_BIAS
        return exponent_16

    @property
    def int_mantissa(self):
        data = self._data
        return (data[1] << 16) | (data[2] << 8) | data[3]

    def __trunc__(self):
        sign = -1 if self.signbit else 1
        exponent_16 = self.exp16
        mantissa = self.int_mantissa

        num_nybbles_to_preserve = min(exponent_16, MAX_BITS_PRECISION_IBM_FLOAT // BITS_PER_NYBBLE)
        num_bits_to_clear = MAX_BITS_PRECISION_IBM_FLOAT - num_nybbles_to_preserve * BITS_PER_NYBBLE
        clear_mask = 2**num_bits_to_clear - 1
        preserve_mask = (2**MAX_BITS_PRECISION_IBM_FLOAT - 1) & ~clear_mask

        truncated_mantissa = mantissa & preserve_mask
        magnitude = int(truncated_mantissa * pow(EXPONENT_BASE, exponent_16)) >> MAX_BITS_PRECISION_IBM_FLOAT
        return sign * magnitude

    def normalize(self):
        """Normalize the floating point representation.

        Returns:
            A normalized IBMFloat equal in value to this object.

        Raises:
            FloatingPointError: If the number could not be normalized.
        """
        if self.is_zero():
            return IBM_FLOAT_ZERO

        exponent_16 = self.exp16
        mantissa = self.int_mantissa

        while mantissa < (1 << 20):
            new_exponent_16 = exponent_16 - 1
            if not (-64 <= new_exponent_16 < 64):
                raise FloatingPointError("Could not normalize {!r} without causing exponent overflow.".format(self))

            mantissa <<= 4
            exponent_16 = new_exponent_16

        exponent_16_biased = exponent_16 + EXPONENT_BIAS

        sign = int(self.signbit) << 7

        a = sign | exponent_16_biased
        b = (mantissa >> 16) & 0xff
        c = (mantissa >> 8) & 0xff
        d = mantissa & 0xff

        return IBMFloat.from_bytes((a, b, c, d))

    def try_normalize(self):
        """Normalize if possible.

        If it is not possible to normalize the representation,
        it remains unmodified.
        """
        try:
            return self.normalize()
        except FloatingPointError:
            return self

    def __floordiv__(self, rhs):
        return floor(float(self) / float(rhs))

    def __rfloordiv__(self, lhs):
        return floor(float(lhs) / float(self))

    def __truediv__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        q = float(self) / float(rhs)
        return IBMFloat.from_float_without_underflow(q) if isinstance(rhs, IBMFloat) else q

    def __rtruediv__(self, lhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        q = float(lhs) / float(self)
        return IBMFloat.from_float_without_underflow(q) if isinstance(lhs, float) else q

    def __pow__(self, exponent):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = pow(float(self), float(exponent))
        if isinstance(p, complex) and p.imag != 0:
            return p
        elif isinstance(exponent, IBMFloat):
            return IBMFloat.from_float_without_underflow(p)
        else:
            return p

    def __rpow__(self, base):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = pow(float(base), float(self))
        if isinstance(p, complex) and p.imag != 0:
            return p
        else:
            return p

    def __mod__(self, rhs):
        """a % b"""
        p = float(self) % float(rhs)
        return IBMFloat.from_float_without_underflow(p) if isinstance(rhs, IBMFloat) else p

    def __rmod__(self, lhs):
        """a % b"""
        return float(lhs) % float(self)

    def __mul__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = float(self) * float(rhs)
        return IBMFloat.from_float_without_underflow(p) if isinstance(rhs, IBMFloat) else p

    def __rmul__(self, lhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = float(lhs) * float(self)
        return IBMFloat.from_float(p) if isinstance(lhs, IBMFloat) else p

    def __add__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = float(self) + float(rhs)
        return IBMFloat.from_float_without_underflow(p) if isinstance(rhs, IBMFloat) else p

    def __radd__(self, lhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(lhs) + float(self)

    def __sub__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        p = float(self) - float(rhs)
        return IBMFloat.from_float_without_underflow(p) if isinstance(rhs, IBMFloat) else p

    def __rsub__(self, lhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(lhs) - float(self)

    def __lt__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(self) < float(rhs)

    def __le__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(self) <= float(rhs)

    def __gt__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(self) > float(rhs)

    def __ge__(self, rhs):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return float(self) >= float(rhs)

    def __ceil__(self):
        t = trunc(self)
        return t if self.signbit else t + 1

    def __floor__(self):
        t = trunc(self)
        return t - 1 if self.signbit else t

    def __round__(self, ndigits=None):
        # Python's 64-bit float has much more precision than this
        # 32-bit IBM float and is much faster, so delegate
        return IBMFloat.from_float_without_underflow(round(float(self), ndigits))

    def __int__(self):
        return trunc(self)


IBM_FLOAT_ZERO = IBMFloat.from_bytes(IBM_ZERO_BYTES)


"""segytools"""

from .segy_header_item import SegyHeaderItem
from .segy_abstract_header import SegyAbstractHeader
from .segy_file_header import SegyFileHeaderRev2
from .segy_trace_header import SegyTraceHeaderRev2


'''
segy abstract header
'''

# import python standard modules
from types import FunctionType, MethodType
from typing import Any

# import 3rd party libraries


# import local python


class SegyAbstractHeader(object):
    """
    Base class for Segy Headers.
    
    SegyAbstractHeader is a container of SegyHeaderItem objects and inherited 
    by SegyFileHeader and SegyTraceHeader. Any custom trace headers classes 
    should inherit this class.

    Methods
    -------
    header_list()
        Returns a list of header item objects.
    key_object_dict()
        Returns a dictionary of header items identifiers and header item objects.
    key_property(name:str, field_id: str)
        Returns a property (field_id) of a segy header item based on the identifier `name`.
    mapped_value(name: str)
        Returns the mapped value of the header item with identifier `name` if the header item value is mapped.
    set_key_property(self, name: str, field_id: str, value: Any)
        Sets the value of the field of a header item based on the identifier `name`.
    """

    def header_list(self) -> list:
        """
        Returns a list of header item objects.
        """
        return [key for key, value in self.__dict__.items() if
                              not isinstance(value, (FunctionType, MethodType))]


    def key_object_dict(self) -> dict:
        """
        Returns a dictionary of header items identifiers and header item objects.
        """
        # stackoverflow : http://stackoverflow.com/questions/21945067/how-list-all-fields-of-an-class-in-python-and-no-methods
        # http://stackoverflow.com/questions/1747817/create-a-dictionary-with-list-comprehension-in-python
        # clever list (of sets) comprehension to get a list of all the fields in the class

        # stack overflow : http://stackoverflow.com/questions/19121722/build-dictionary-in-python-loop-list-and-dictionary-comprehensions
        # clever dict comprehension to get a dictionary of all the field.values in a class
        # {_key: _value(_key) for _key in _container}

        SETLIST_HDR_OBJECT = [(key, obj) for key, obj in self.__dict__.items() if
                              not isinstance(obj, (FunctionType, MethodType))]
        #print(len(SETLIST_HDR_OBJECT))

        # return {name: val for (name, val) in self.header_list()}
        return {name: val for (name, val) in SETLIST_HDR_OBJECT}

    def key_property(self, name: str, field_id: str) -> Any:
        """
        Returns a property (field_id) of a 'SegyHeaderItem' based on the identifier `name`.

        Parameters
        ----------
        name : str
            The header item identifier.
        field_id : str
            The field or property of the `SegyHeaderItem` to be retrived.
        """
        # obj is a SegyHeaderItem object
        obj = self.key_object_dict()[name]
        return obj.field(field_id)

    def mapped_value(self, name: str) -> Any:
        """
        Returns the mapped value of the header item with identifier `name` if the header item value is mapped.

        Parameters
        ----------
        name : str
            The `SegyHeaderItem` identifier.
        """
        # get the value from the map_dict if the value from the object (in this case, the key) exists, else
        # return the original value
        # obj is a SegyHeaderTemplate object
        obj = self.key_object_dict()[name]
        return obj.map_dict[obj.value] if obj.value in obj.map_dict else obj.value

    def set_key_property(self, name: str, field_id: str, value: Any) -> None:
        """
        Sets the value of the field of a header item based on the identifier `name`.

        Parameters
        ----------
        name : str
            The `SegyHeaderItem` identifier.
        field_id : str
            The field or property of the `SegyHeaderItem` to be retrived.
        value : Any
            The value that will be assigned to the `SegyHeaderItem` field.
        """
        # obj is a SegyHeaderTemplate object
        obj = self.key_object_dict()[name]
        obj.set_field(field_id, value)

    def _to_bytes(self, endianess: str, byte_length: int = 240) -> bytes:
        """
        Converts each header item to a byte object and returns the complete 
        bytes object of length `byte_length`.

        Parameters
        ----------
        endianess : str
            Either 'big' or 'little'
        byte_length : int
            the number of bytes in the output bytes object. default is 240.
        """
        bsgy = bytearray(byte_length)
        file_key_obj_dict = self.key_object_dict()
        for _, obj in file_key_obj_dict.items():
            
            if obj.map_bool is True:
                # reverse map back to integer
                # tmp_int = list(obj.map_dict.keys())[list(obj.map_dict.values()).index(obj.value)]
                tmp_int = 0
                for key, val in obj.map_dict.items():
                    if obj.value == val:
                        tmp_int = int(key)
                        break
                obj.value = tmp_int
            tmp = int.to_bytes(int(obj.value), length=obj.nbytes, byteorder=endianess, signed=obj.signed)
            bsgy[obj.startbyte - 1:obj.startbyte + obj.nbytes - 1] = tmp
        return bytes(bsgy)


"""
XPS parser
"""
import numpy
from pandas import DataFrame, Series

# H26
# H26      Relation Record Specification
# H26
# H26 DEFINITION OF FIELD            COLS       FORMAT       UNIT
# H26 Record identification          1- 1           A1       none
# H26 Field tape number (l adj)      2- 7          3A2       none
# H26 Field record number            8-11           I4       none
# H26 Field record increment        12-12           I1       none
# H26 Instrument code               13-13           A1       none
# H26 Line name (left adj)          14-29          4A4       none
# H26 Point number (right adj)      30-37          2A4       none
# H26 Point index                   38-38           I1       none
# H26 From channel                  39-42           I4       none
# H26 To channel                    43-46           I4       none
# H26 Channel increment             47-47           I1       none
# H26 Line name (left adj)          48-63          4A4       none
# H26 From receiver (right adj)     64-71          2A4       none
# H26 To receiver (right adj)       72-79          2A4       none
# H26 Receiver index                80-80           A1       none
# H26

class XPSLineItem(object):
    def __init__(self, s: str):
        self.field_tape_number = s[1:7].strip()
        self.field_record_number = s[7:11].strip()
        self.field_record_increment = s[11:12].strip()
        self.instrument_code = s[12:13].strip()
        self.sline_name = s[13:29].strip()
        self.point_number = s[29:37].strip()
        self.point_index = s[37:38].strip()
        self.from_channel = s[38:42].strip()
        self.to_channel = s[42:46].strip()
        self.channel_increment = s[46:47].strip()
        self.rline_name = s[47:63].strip()
        self.from_receiver = s[63:71].strip()
        self.to_receiver = s[71:79].strip()
        self.receiver_index = s[79:80].strip()

    def __str__(self):
        srep = '\nsource receiver spread\n'
        srep += 'Field tape number      : ' + self.field_tape_number + '\n'
        srep += 'Field record number    : ' + self.field_record_number + '\n'
        srep += 'Field record increment : ' + self.field_record_increment + '\n'
        srep += 'Instrument code        : ' + self.instrument_code + '\n'
        srep += 'Source line name       : ' + self.sline_name + '\n'
        srep += 'Point number           : ' + self.point_number + '\n'
        srep += 'Point index            : ' + self.point_index + '\n'
        srep += 'From channel           : ' + self.from_channel + '\n'
        srep += 'To channel             : ' + self.to_channel + '\n'
        srep += 'Channel increment      : ' + self.channel_increment + '\n'
        srep += 'Receiver line name     : ' + self.rline_name + '\n'
        srep += 'From receiver          : ' + self.from_receiver + '\n'
        srep += 'To receiver            : ' + self.to_receiver + '\n'
        srep += 'Receiver index         : ' + self.receiver_index
        return srep

class XPS(object):
    def __init__(self, f: str):
        field_tape_numbers = []
        field_record_numbers = []
        field_record_increments = []
        instrument_codes = []
        sline_names = []
        point_numbers = []
        point_indices = []
        from_channels = []
        to_channels = []
        channel_increments = []
        rline_names = []
        from_receivers = []
        to_receivers = []
        receiver_indices = []
        with open(f, "r") as fobj:
            for cnt, line in enumerate(fobj):
                if not line.startswith("H26"):
                    record = XPSLineItem(line)
                    field_tape_numbers.append(int(record.field_tape_number))
                    field_record_numbers.append(int(record.field_record_number))
                    field_record_increments.append(int(record.field_record_increment))
                    instrument_codes.append(record.instrument_code)
                    sline_names.append(record.sline_name)
                    point_numbers.append(int(record.point_number))
                    point_indices.append(int(record.point_index))
                    from_channels.append(int(record.from_channel))
                    to_channels.append(int(record.to_channel))
                    channel_increments.append(int(record.channel_increment))
                    rline_names.append(record.rline_name)
                    from_receivers.append(int(record.from_receiver))
                    to_receivers.append(int(record.to_receiver))
                    receiver_indices.append(int(record.receiver_index))

            fobj.close()
        # temporary dict for dataframe constructor
        tmp_dict = dict()
        tmp_dict['field_tape_number'] = field_tape_numbers
        tmp_dict['field_record_number'] = field_record_numbers
        tmp_dict['field_record_increment'] = field_record_increments
        tmp_dict['instrument_code'] = instrument_codes
        tmp_dict['sline_name'] = sline_names
        tmp_dict['point_number'] = point_numbers
        tmp_dict['point_index'] = point_indices
        tmp_dict['from_channel'] = from_channels
        tmp_dict['to_channel'] = to_channels
        tmp_dict['channel_increment'] = channel_increments
        tmp_dict['rline_name'] = rline_names
        tmp_dict['from_receiver'] = from_receivers
        tmp_dict['to_receiver'] = to_receivers
        tmp_dict['receiver_index'] = receiver_indices
        self._df = DataFrame(tmp_dict)

    def dataframe(self):
        return self._df

    def source_point(self, ffid: int):
        tmpi = self._df.loc[self._df['field_record_number'] == ffid].index.values
        return self._df.loc[tmpi[0], 'point_number']

    def reciever_station(self, ffid: int, chan: int):
        tmpi = self._df.loc[self._df['field_record_number'] == ffid].index.values
        first_chan = self._df.loc[tmpi[0], 'from_channel']
        last_chan = self._df.loc[tmpi[0], 'to_channel']
        first_rec = self._df.loc[tmpi[0], 'from_receiver']
        last_rec = self._df.loc[tmpi[0], 'to_receiver']
        chan_linspace = numpy.linspace(start=first_chan, stop=last_chan, num=last_chan-first_chan+1, endpoint=True, dtype=numpy.dtype(int))
        recs_linspace = numpy.linspace(start=first_rec, stop=last_rec, num=last_chan-first_chan+1, endpoint=True, dtype=numpy.dtype(int))
        recstation = recs_linspace[numpy.nonzero(chan_linspace == chan)]
        if not recstation.size == 0:
            rvalue = recstation[0]
        else:
            rvalue = 0
        return rvalue

    def receiver_stations(self, ffid: int):
        tmpi = self._df.loc[self._df['field_record_number'] == ffid].index.values
        first_chan = self._df.loc[tmpi[0], 'from_channel']
        last_chan = self._df.loc[tmpi[0], 'to_channel']
        first_rec = self._df.loc[tmpi[0], 'from_receiver']
        last_rec = self._df.loc[tmpi[0], 'to_receiver']
        chan_linspace = numpy.linspace(start=first_chan, stop=last_chan, num=last_chan - first_chan + 1, endpoint=True, dtype=numpy.dtype(int))
        recs_linspace = numpy.linspace(start=first_rec, stop=last_rec, num=last_chan - first_chan + 1, endpoint=True, dtype=numpy.dtype(int))
        sr = Series(data=recs_linspace, index=chan_linspace)
        return sr

# END


'''
segy trace header
'''

# import python standard modules


# import 3rd party libraries


# import local python
from .segy_abstract_header import SegyAbstractHeader
from .segy_header_item import SegyHeaderItem


# bytes 29-30
TRACE_IDENTIFICATION_CODE = {
    -1: 'Other',
    0: 'Unknown',
    1: 'Seismic Data',
    2: 'Dead',
    3: 'Dummy',
    4: 'Time break',
    5: 'Uphole',
    6: 'Sweep',
    7: 'Timing',
    8: 'Waterbreak',
    9: 'Near-field gun signature',
    10: 'Far-field gun signature',
    11: 'Seismic pressure sensor',
    12: 'Multi-component seismic sensor - Vertical component',
    13: 'Multi-component seismic sensor - Cross-line component',
    14: 'Multi-component seismic sensor - In-line component',
    15: 'Rotated multi-component seismic sensor - Vertical component',
    16: 'Rotated multi-component seismic sensor - Transverse component',
    17: 'Rotated multi-component seismic sensor - Radial component',
    18: 'Vibrator reaction mass',
    19: 'Vibrator baseplate',
    20: 'Vibrator estimated ground force',
    21: 'Vibrator reference',
    22: 'Time-Velocity Pairs'
}

# bytes 35-36
DATA_USE = {
    1: 'Production',
    2: 'Test'
}

# bytes 69-70 and 71-72
COORDINATE_SCALAR_MULTIPLIER = {
    -10000: 0.0001,
    -1000: 0.001,
    -100: 0.01,
    -10: 0.1,
    -1: 1,
    1: 1,
    10: 10,
    100: 100,
    1000: 1000,
    10000: 10000
}

# bytes 89-90
COORDINATE_UNITS = {
    1: 'Length (meters or feet)',
    2: 'Seconds or arc',
    3: 'Decimal degrees',
    4: 'Degrees, minutes, seconds (DMS)'
}
# TODO: properly handle anything other than length; most routines are based on Cartesian system

# bytes 119-120
GAIN_TYPE_OF_FIELD_INSTRUMENTS = {
    1: 'fixed',
    2: 'binary',
    3: 'floating point',
    4: 'optional use'
}

# bytes 125-126
CORRELATED = {
    1: 'yes',
    2: 'no'
}

# bytes 133-134
SWEEP_TYPE = {
    1: 'linear',
    2: 'parabolic',
    3: 'exponential',
    4: 'other'
}

# bytes 139-140
TAPER_TYPE = {
    1: 'linear',
    2: 'cos**2',
    3: 'other'
}

# bytes 167-168
TIME_BASIS_CODE = {
    1: 'Local',
    2: 'GMT (Greenwich Mean Time)',
    3: 'Other',
    4: 'UTC (Coordinated Universal Time)'
}

# bytes 179-180
OVER_TRAVEL_ASSOCIATED_TAPER_AT_BEGINNING_OR_END_OF_LINE = {
    1: 'down (or behind)',
    2: 'up (or ahead)'
}

# bytes 203-204
TRACE_VALUE_MEASUREMENT_UNIT = {
    -1: 'Other',
    0: 'Unknown',
    1: 'Pascal (Pa)',
    2: 'Volts (v)',
    3: 'Millivolts (mV)',
    4: 'Amperes (A)',
    5: 'Meters (m)',
    6: 'Meters per second (m/s)',
    7: 'Meters per second squared (m/(s**2))',
    8: 'Newton (N)',
    9: 'Watt (W)'
}

# bytes 211-212
TRANSDUCTION_UNITS = {
    -1: 'Other',
    0: 'Unknown',
    1: 'Pascal (Pa)',
    2: 'Volts (v)',
    3: 'Millivolts (mV)',
    4: 'Amperes (A)',
    5: 'Meters (m)',
    6: 'Meters per second (m/s)',
    7: 'Meters per second squared (m/(s**2))',
    8: 'Newton (N)',
    9: 'Watt (W)'
}

# bytes 217-218
SOURCE_TYPE_ORIENTATION ={
    -1: 'Other',
    0: 'Unknown',
    1: 'Vibratory - Vertical orientation',
    2: 'Vibratory - Cross-line orientation',
    3: 'Vibratory - In-line-orientation',
    4: 'Impulsive - Vertical orientation',
    5: 'Impulsive - Cross-line orientation',
    6: 'Impulsive - In-line-orientation',
    7: 'Distributed Impulsive - Vertical orientation',
    8: 'Distributed Impulsive - Cross-line orientation',
    9: 'Distributed Impulsive - In-line-orientation',
}

# bytes 231-232
SOURCE_MEASUREMENT_UNIT = {
    -1 : 'Other',
    0: 'Unknown',
    1: 'Joule (J)',
    2: 'Kilowatt (kW)',
    3: 'Pascal (P)',
    4: 'Bar (Bar)',
    5: 'Newton (N)',
    6: 'Kilograms (kg)'
}

class SegyTraceHeaderRev2(SegyAbstractHeader):

    def __init__(self):
        super().__init__()

        # HEADERS
        self.trseql = SegyHeaderItem(
            name='trseql', 
            nbytes=4, 
            startbyte=1,
            description='trace_sequence_number_within_line',
            signed=True)
        self.trseqf = SegyHeaderItem(
            name='trseqf', 
            nbytes=4, 
            startbyte=5,
            description='trace_sequence_number_within_segy_file',
            signed=True)
        self.ffid = SegyHeaderItem(
            name='ffid', 
            nbytes=4, 
            startbyte=9,
            description='original_field_record_number',
            signed=True)
        self.trffid = SegyHeaderItem(
            name='trffid', 
            nbytes=4, 
            startbyte=13,
            description='trace_number_within_the_original_field_record',
            signed=True)
        self.shot = SegyHeaderItem(
            name='shot', 
            nbytes=4, 
            startbyte=17,
            description='energy_source_point_number',
            signed=True)
        self.ens = SegyHeaderItem(
            name='ens', 
            nbytes=4, 
            startbyte=21,
            description='ensemble_number',
            signed=True)
        self.trcens = SegyHeaderItem(
            name='trcens', 
            nbytes=4, 
            startbyte=25,
            description='trace_number_within_the_ensemble',
            signed=True)
        self.trid = SegyHeaderItem(
            name='trid', 
            nbytes=2, 
            startbyte=29,
            description='trace_identification_code',
            signed=True,
            map_bool=True, 
            map_dict=TRACE_IDENTIFICATION_CODE)
        self.vsum = SegyHeaderItem(
            name='vsum', 
            nbytes=2, 
            startbyte=31,
            description='number_of_vertically_summed_traces_yielding_this_trace',
            signed=True)
        self.hsum = SegyHeaderItem(
            name='hsum', 
            nbytes=2, 
            startbyte=33,
            description='number_of_horizontally_stacked_traces_yielding_this_trace',
            signed=True)
        self.datuse = SegyHeaderItem(
            name='datuse', 
            nbytes=2, 
            startbyte=35,
            description='data_use',
            signed=True,
            map_bool=True, 
            map_dict=DATA_USE)
        self.offset = SegyHeaderItem(
            name='offset', 
            nbytes=4, 
            startbyte=37,
            description='distance_from_center_of_the_source_point_to_the_center_of_the_receiver_group',
            signed=True)
        self.recelev = SegyHeaderItem(
            name='recelev', 
            nbytes=4, 
            startbyte=41,
            description='receiver_group_elevation',
            signed=True)
        self.srcelev = SegyHeaderItem(
            name='srcelev', 
            nbytes=4, 
            startbyte=45,
            description='surface_elevation_at_source',
            signed=True)
        self.srcdepth = SegyHeaderItem(
            name='srcdepth', 
            nbytes=4, 
            startbyte=49,
            description='source_depth_below_surface',
            signed=True)
        self.datmrec = SegyHeaderItem(
            name='datmrec', 
            nbytes=4, 
            startbyte=53,
            description='datum_elevation_at_receiver_group',
            signed=True)
        self.datmsrc = SegyHeaderItem(
            name='datmsrc', 
            nbytes=4, 
            startbyte=57,
            description='datum_elevation_at_source',
            signed=True)
        self.h2ozsrc = SegyHeaderItem(
            name='h2ozsrc', 
            nbytes=4, 
            startbyte=61,
            description='water_depth_at_source',
            signed=True)
        self.h2ozrec = SegyHeaderItem(
            name='h2ozrec', 
            nbytes=4, 
            startbyte=65,
            description='water_depth_at_group',
            signed=True)
        self.zsclr = SegyHeaderItem(
            name='zsclr', 
            nbytes=2, 
            startbyte=69,
            description='scalar_to_be_applied_to_all_elevations_and_depths',
            signed=True,
            map_bool=True, 
            map_dict=COORDINATE_SCALAR_MULTIPLIER)
        self.xysclr = SegyHeaderItem(
            name='xysclr', 
            nbytes=2, 
            startbyte=71,
            description='scalar_to_be_applied_to_all_coordinates',
            signed=True,
            map_bool=True, 
            map_dict=COORDINATE_SCALAR_MULTIPLIER)
        self.srcx = SegyHeaderItem(
            name='srcx', 
            nbytes=4, 
            startbyte=73,
            description='source_coordinate_x',
            signed=True)
        self.srcy = SegyHeaderItem(
            name='srcy', 
            nbytes=4, 
            startbyte=77,
            description='source_coordinate_y',
            signed=True)
        self.recx = SegyHeaderItem(
            name='recx', 
            nbytes=4, 
            startbyte=81,
            description='group_coordinate_x',
            signed=True)
        self.recy = SegyHeaderItem(
            name='recy', 
            nbytes=4, 
            startbyte=85,
            description='group_coordinate_y',
            signed=True)
        self.coordunits = SegyHeaderItem(
            name='coordunits', 
            nbytes=2, 
            startbyte=89,
            description='coordinate_units',
            signed=True,
            map_bool=True, 
            map_dict=COORDINATE_UNITS)
        self.weathvel = SegyHeaderItem(
            name='weathvel', 
            nbytes=2, 
            startbyte=91,
            description='weathering_velocity',
            signed=True)
        self.subweathvel = SegyHeaderItem(
            name='subweathvel', 
            nbytes=2, 
            startbyte=93,
            description='subweathering_velocity',
            signed=True)
        self.uptsrc = SegyHeaderItem(
            name='uptsrc', 
            nbytes=2, 
            startbyte=95,
            description='uphole_time_at_source_in_ms',
            signed=True)
        self.uptrec = SegyHeaderItem(
            name='uptrec', 
            nbytes=2, 
            startbyte=97,
            description='uphole_time_at_group_in_ms',
            signed=True)
        self.srcstat = SegyHeaderItem(
            name='srcstat', 
            nbytes=2, 
            startbyte=99,
            description='source_static_correction_in_ms',
            signed=True)
        self.recstat = SegyHeaderItem(
            name='recstat', 
            nbytes=2, 
            startbyte=101,
            description='group_static_correction_in_ms',
            signed=True)
        self.totstat = SegyHeaderItem(
            name='totstat', 
            nbytes=2, 
            startbyte=103,
            description='total_static_applied_in_ms',
            signed=True)
        self.lagA = SegyHeaderItem(
            name='lagA', 
            nbytes=2, 
            startbyte=105,
            description='lag_time_A',
            signed=True)
        self.lagB = SegyHeaderItem(
            name='lagB', 
            nbytes=2, 
            startbyte=107,
            description='lag_time_B',
            signed=True)
        self.delayt = SegyHeaderItem(
            name='delayt', 
            nbytes=2, 
            startbyte=109,
            description='delay_recording_time',
            signed=True)
        self.mutes = SegyHeaderItem(
            name='mutes', 
            nbytes=2, 
            startbyte=111,
            description='mute_time_start_time_in_ms',
            signed=True)
        self.mutee = SegyHeaderItem(
            name='mutee', 
            nbytes=2, 
            startbyte=113,
            description='mute_time_end_time_in_ms',
            signed=True)
        self.numsmp = SegyHeaderItem(
            name='numsmp', 
            nbytes=2, 
            startbyte=115,
            description='number_of_samples_in_this_trace',
            signed=True)  # verify this matches file header
        self.smpint = SegyHeaderItem(
            name='smpint', 
            nbytes=2, 
            startbyte=117,
            description='sample_interval_in_ms_for_this_trace',
            signed=True) # verify this matches file header
        self.gaintype = SegyHeaderItem(
            name='gaintype', 
            nbytes=2, 
            startbyte=119,
            description='gain_type_of_field_instruments',
            signed=True,
            map_bool=True, 
            map_dict=GAIN_TYPE_OF_FIELD_INSTRUMENTS)
        self.gainconst = SegyHeaderItem(
            name='gainconst', 
            nbytes=2, 
            startbyte=121,
            description='instrument_gain_constant',
            signed=True)
        self.initgain = SegyHeaderItem(
            name='initgain', 
            nbytes=2, 
            startbyte=123,
            description='instrument_early_or_initial_gain',
            signed=True)
        self.corrx = SegyHeaderItem(
            name='corrx', 
            nbytes=2, 
            startbyte=125,
            description='correlated',
            signed=True,
            map_bool=True, 
            map_dict=CORRELATED)
        self.sweepfs = SegyHeaderItem(
            name='sweepfs', 
            nbytes=2, 
            startbyte=127,
            description='sweep_frequency_at_start',
            signed=True)
        self.sweepfe = SegyHeaderItem(
            name='sweepfe', 
            nbytes=2, 
            startbyte=129,
            description='sweep_frequency_at_end',
            signed=True)
        self.sweeplen = SegyHeaderItem(
            name='sweeplen', 
            nbytes=2, 
            startbyte=131,
            description='sweep_length_in_ms',
            signed=True)
        self.sweeptype = SegyHeaderItem(
            name='sweeptype', 
            nbytes=2, 
            startbyte=133,
            description='sweep_type',
            signed=True,
            map_bool=True, 
            map_dict=SWEEP_TYPE)
        self.sweeptprs = SegyHeaderItem(
            name='sweeptprs', 
            nbytes=2, 
            startbyte=135,
            description='sweep_trace_taper_length_at_start_in_ms',
            signed=True)
        self.sweeptpre = SegyHeaderItem(
            name='sweeptpre', 
            nbytes=2, 
            startbyte=137,
            description='sweep_trace_taper_length_at_end_in_ms',
            signed=True)
        self.tprtype = SegyHeaderItem(
            name='tprtype', 
            nbytes=2, 
            startbyte=139,
            description='taper_type',
            signed=True,
            map_bool=True, 
            map_dict=TAPER_TYPE)
        self.aliasfilf = SegyHeaderItem(
            name='aliasfilf', 
            nbytes=2, 
            startbyte=141,
            description='alias_filter_frequency',
            signed=True)
        self.aliasfils = SegyHeaderItem(
            name='aliasfils', 
            nbytes=2, 
            startbyte=143,
            description='alias_filter_slope',
            signed=True)
        self.notchfilf = SegyHeaderItem(
            name='notchfilf', 
            nbytes=2, 
            startbyte=145,
            description='notch_filter_frequency',
            signed=True)
        self.notchfils = SegyHeaderItem(
            name='notchfils', 
            nbytes=2, 
            startbyte=147,
            description='notch_filter_slope',
            signed=True)
        self.lowcutf = SegyHeaderItem(
            name='lowcutf', 
            nbytes=2, 
            startbyte=149,
            description='low_cut_frequency',
            signed=True)
        self.highcutf = SegyHeaderItem(
            name='highcutf', 
            nbytes=2, 
            startbyte=151,
            description='high_cut_frequency',
            signed=True)
        self.lowcuts = SegyHeaderItem(
            name='lowcuts', 
            nbytes=2, 
            startbyte=153,
            description='low_cut_slope',
            signed=True)
        self.highcuts = SegyHeaderItem(
            name='highcuts', 
            nbytes=2, 
            startbyte=155,
            description='high_cut_slope',
            signed=True)
        self.year = SegyHeaderItem(
            name='year', 
            nbytes=2, 
            startbyte=157,
            description='year_data_recorded',
            signed=True)
        self.day = SegyHeaderItem(
            name='day', 
            nbytes=2, 
            startbyte=159,
            description='day_of_year',
            signed=True)
        self.hour = SegyHeaderItem(
            name='hour', 
            nbytes=2, 
            startbyte=161,
            description='hour_of_day',
            signed=True)
        self.minute = SegyHeaderItem(
            name='minute', 
            nbytes=2, 
            startbyte=163,
            description='minute_of_hour',
            signed=True)
        self.second = SegyHeaderItem(
            name='second', 
            nbytes=2, 
            startbyte=165,
            description='second_of_minute',
            signed=True)
        self.timebasis = SegyHeaderItem(
            name='timebasis', 
            nbytes=2, 
            startbyte=167,
            description='time_basis_code',
            signed=True,
            map_bool=True, 
            map_dict=TIME_BASIS_CODE)
        self.trcweight = SegyHeaderItem(
            name='trcweight', 
            nbytes=2, 
            startbyte=169,
            description='trace_weighting_factor',
            signed=True)
        self.groupnum = SegyHeaderItem(
            name='groupnum', 
            nbytes=2, 
            startbyte=171,
            description='geophone_group_number_of_roll_switch_position_one',
            signed=True)
        self.groupnumfirst = SegyHeaderItem(
            name='groupnumfirst', 
            nbytes=2, 
            startbyte=173,
            description='geophone_group_number_of_trace_number_one',
            signed=True)
        self.groupnumlast = SegyHeaderItem(
            name='groupnumlast', 
            nbytes=2, 
            startbyte=175,
            description='geophone_group_number_of_last_trace',
            signed=True)
        self.gapsize = SegyHeaderItem(
            name='gapsize', 
            nbytes=2, 
            startbyte=177,
            description='gap_size',
            signed=True)
        self.overtrvl = SegyHeaderItem(
            name='overtrvl', 
            nbytes=2, 
            startbyte=179,
            description='over_travel_associated_with_taper',
            signed=True,
            map_bool=True, 
            map_dict=OVER_TRAVEL_ASSOCIATED_TAPER_AT_BEGINNING_OR_END_OF_LINE)
        self.ensx = SegyHeaderItem(
            name='ensx', 
            nbytes=4, 
            startbyte=181,
            description='x_coordinate_of_ensemble_position_of_this_trace',
            signed=True)
        self.ensy = SegyHeaderItem(
            name='ensy', 
            nbytes=4, 
            startbyte=185,
            description='y_coordinate_of_ensemble_position_of_this_trace',
            signed=True)
        self.il = SegyHeaderItem(
            name='il', 
            nbytes=4, 
            startbyte=189,
            description='for_3d_poststack_data_this_field_is_for_in_line_number',
            signed=True)
        self.xl = SegyHeaderItem(
            name='xl', 
            nbytes=4, 
            startbyte=193,
            description='for_3d_poststack_data_this_field_is_for_cross_line_number',
            signed=True)
        self.shotpoint = SegyHeaderItem(
            name='shotpoint', 
            nbytes=4, 
            startbyte=197,
            description='shotpoint_number',
            signed=True)
        self.shotptsclr = SegyHeaderItem(
            name='shotptsclr', 
            nbytes=2, 
            startbyte=201,
            description='scalar_to_be_applied_to_the_shotpoint_number',
            signed=True)
        self.trcvalmeas = SegyHeaderItem(
            name='trcvalmeas', 
            nbytes=2, 
            startbyte=203,
            description='trace_value_measurement_unit',
            signed=True,
            map_bool=True, 
            map_dict=TRACE_VALUE_MEASUREMENT_UNIT)
        # The transduction constant is encoded with the mantissa and the power of
        # the exponent, e.g.:
        # transduction_constant =
        # transduction_constant_mantissa * 10 ** transduction_constant_exponent
        self.transconstm = SegyHeaderItem(
            name='transconstm', 
            nbytes=4, 
            startbyte=205,
            description='transduction_constant_mantissa',
            signed=True)
        self.transconste = SegyHeaderItem(
            name='transconste', 
            nbytes=2, 
            startbyte=209,
            description='transduction_constant_exponent',
            signed=True)
        self.transunits = SegyHeaderItem(
            name='transunits', 
            nbytes=2, 
            startbyte=211,
            description='transduction_units',
            signed=True,
            map_bool=True, 
            map_dict=TRANSDUCTION_UNITS)
        self.deviceid = SegyHeaderItem(
            name='deviceid', 
            nbytes=2, 
            startbyte=213,
            description='device_trace_identifier',
            signed=True)
        self.timesclr = SegyHeaderItem(
            name='timesclr', 
            nbytes=2, 
            startbyte=215,
            description='scalar_to_be_applied_to_times',
            signed=True)
        self.srctypeor = SegyHeaderItem(
            name='srctypeor', 
            nbytes=2, 
            startbyte=217,
            description='source_type_orientation',
            signed=True,
            map_bool=True, 
            map_dict=SOURCE_TYPE_ORIENTATION)
        self.srcdirm = SegyHeaderItem(
            name='srcdirm', 
            nbytes=4, 
            startbyte=219,
            description='source_energy_direction_mantissa',
            signed=True)
        self.srcdire = SegyHeaderItem(
            name='srcdire', 
            nbytes=2, 
            startbyte=223,
            description='source_energy_direction_exponent',
            signed=True)
        # The source measurement is encoded with the mantissa and the power of
        # the exponent, e.g.:
        # source_measurement =
        # source_measurement_mantissa * 10 ** source_measurement_exponent
        self.srcmeasm = SegyHeaderItem(
            name='srcmeasm', 
            nbytes=4, 
            startbyte=225,
            description='source_measurement_mantissa',
            signed=True)
        self.srcmease = SegyHeaderItem(
            name='srcmease', 
            nbytes=2, 
            startbyte=229,
            description='source_measurement_exponent',
            signed=True)
        self.srcmeasu = SegyHeaderItem(
            name='srcmeasu', 
            nbytes=2, 
            startbyte=231,
            description='source_measurement_unit',
            signed=True,
            map_bool=True, 
            map_dict=SOURCE_MEASUREMENT_UNIT)
        self.unknown233 = SegyHeaderItem(
            name='unknown233', 
            nbytes=4, 
            startbyte=233,
            description='unknown',
            signed=True)
        self.unknown237 = SegyHeaderItem(
            name='unknown237', 
            nbytes=4, 
            startbyte=237,
            description='unknown',
            signed=True)

    def __str__(self):
        s = '\nsegy trace header values\n'
        s += str(self.trseql) + '\n'
        s += str(self.trseqf) + '\n'
        s += str(self.ffid) + '\n'
        s += str(self.trffid) + '\n'
        s += str(self.shot) + '\n'
        s += str(self.ens) + '\n'
        s += str(self.trcens) + '\n'
        s += str(self.trid) + '\n'
        s += str(self.vsum) + '\n'
        s += str(self.hsum) + '\n'
        s += str(self.datuse) + '\n'
        s += str(self.offset) + '\n'
        s += str(self.recelev) + '\n'
        s += str(self.srcelev) + '\n'
        s += str(self.srcdepth) + '\n'
        s += str(self.datmrec) + '\n'
        s += str(self.datmsrc) + '\n'
        s += str(self.h2ozsrc) + '\n'
        s += str(self.h2ozrec) + '\n'
        s += str(self.zsclr) + '\n'
        s += str(self.xysclr) + '\n'
        s += str(self.srcx) + '\n'
        s += str(self.srcy) + '\n'
        s += str(self.recx) + '\n'
        s += str(self.recy) + '\n'
        s += str(self.coordunits) + '\n'
        s += str(self.weathvel) + '\n'
        s += str(self.subweathvel) + '\n'
        s += str(self.uptsrc) + '\n'
        s += str(self.uptrec) + '\n'
        s += str(self.srcstat) + '\n'
        s += str(self.recstat) + '\n'
        s += str(self.totstat) + '\n'
        s += str(self.lagA) + '\n'
        s += str(self.lagB) + '\n'
        s += str(self.delayt) + '\n'
        s += str(self.mutes) + '\n'
        s += str(self.mutee) + '\n'
        s += str(self.numsmp) + '\n'
        s += str(self.smpint) + '\n'
        s += str(self.gaintype) + '\n'
        s += str(self.gainconst) + '\n'
        s += str(self.initgain) + '\n'
        s += str(self.corrx) + '\n'
        s += str(self.sweepfs) + '\n'
        s += str(self.sweepfe) + '\n'
        s += str(self.sweeplen) + '\n'
        s += str(self.sweeptype) + '\n'
        s += str(self.sweeptprs) + '\n'
        s += str(self.sweeptpre) + '\n'
        s += str(self.tprtype) + '\n'
        s += str(self.aliasfilf) + '\n'
        s += str(self.aliasfils) + '\n'
        s += str(self.notchfilf) + '\n'
        s += str(self.notchfils) + '\n'
        s += str(self.lowcutf) + '\n'
        s += str(self.highcutf) + '\n'
        s += str(self.lowcuts) + '\n'
        s += str(self.highcuts) + '\n'
        s += str(self.year) + '\n'
        s += str(self.day) + '\n'
        s += str(self.hour) + '\n'
        s += str(self.minute) + '\n'
        s += str(self.second) + '\n'
        s += str(self.timebasis) + '\n'
        s += str(self.trcweight) + '\n'
        s += str(self.groupnum) + '\n'
        s += str(self.groupnumfirst) + '\n'
        s += str(self.groupnumlast) + '\n'
        s += str(self.gapsize) + '\n'
        s += str(self.overtrvl) + '\n'
        s += str(self.ensx) + '\n'
        s += str(self.ensy) + '\n'
        s += str(self.il) + '\n'
        s += str(self.xl) + '\n'
        s += str(self.shotpoint) + '\n'
        s += str(self.shotptsclr) + '\n'
        s += str(self.trcvalmeas) + '\n'
        s += str(self.transconstm) + '\n'
        s += str(self.transconste) + '\n'
        s += str(self.transunits) + '\n'
        s += str(self.deviceid) + '\n'
        s += str(self.timesclr) + '\n'
        s += str(self.srctypeor) + '\n'
        s += str(self.srcdirm) + '\n'
        s += str(self.srcdire) + '\n'
        s += str(self.srcmeasm) + '\n'
        s += str(self.srcmease) + '\n'
        s += str(self.srcmeasu) + '\n'
        s += str(self.unknown233) + '\n'
        s += str(self.unknown237)
        return s


    # --- METHODS ---
    def set_scalar_corrected_cooridinates(self):
        if self.zsclr.value != 1:
            sclr = self.zsclr.value
            self.recelev.value = self.recelev.value * sclr
            self.srcelev.value = self.srcelev.value * sclr
            self.srcdepth.value = self.srcdepth.value * sclr
            self.datmsrc.value = self.datmsrc.value * sclr
            self.datmrec.value = self.datmrec.value * sclr
            self.h2ozsrc.value = self.h2ozsrc.value * sclr
            self.h2ozrec.value = self.h2ozrec.value * sclr

        if self.xysclr.value != 1:
            sclr = self.xysclr.value
            self.srcx.value = self.srcx.value * sclr
            self.srcy.value = self.srcy.value * sclr
            self.recx.value = self.recx.value * sclr
            self.recy.value = self.recy.value * sclr

    def set_trace_header_values(self, bsgy, endianess):
        # bsgy is type bytes
        file_key_obj_dict = self.key_object_dict()
        for name, obj in file_key_obj_dict.items():
            bobj = bsgy[obj.startbyte - 1:obj.startbyte + obj.nbytes - 1]
            obj.value = int.from_bytes(bobj, byteorder=endianess, signed=obj.signed)

            if obj.map_bool is True:
                self.set_key_property(obj.name, 'value', self.mapped_value(obj.name))

    def to_bytes(self, endianess, byte_length=240):
        # bsgy = bytearray(byte_length)
        # file_key_obj_dict = self.key_object_dict()
        # reverse scaler coordinates for write
        # if self.zsclr.value != 0:
        #     self.recelev.value = self.recelev.value / self.zsclr.value
        #     self.srcelev.value = self.srcelev.value / self.zsclr.value
        #     self.srcdepth.value = self.srcdepth.value / self.zsclr.value
        #     self.datmsrc.value = self.datmsrc.value / self.zsclr.value
        #     self.datmrec.value = self.datmrec.value / self.zsclr.value
        #     self.h2ozsrc.value = self.h2ozsrc.value / self.zsclr.value
        #     self.h2ozrec.value = self.h2ozrec.value / self.zsclr.value
        # if self.xysclr.value != 0:
        #     self.srcx.value = self.srcx.value / self.xysclr.value
        #     self.srcy.value = self.srcy.value / self.xysclr.value
        #     self.recx.value = self.recx.value / self.xysclr.value
        #     self.recy.value = self.recy.value / self.xysclr.value
        # for name, obj in file_key_obj_dict.items():
            
        #     if obj.map_bool is True:
        #         # reverse map back to integer
        #         # tmp_int = list(obj.map_dict.keys())[list(obj.map_dict.values()).index(obj.value)]
        #         tmp_int = 0
        #         for key, val in obj.map_dict.items():
        #             if obj.value == val:
        #                 tmp_int = int(key)
        #                 break
        #         obj.value = tmp_int
        #     tmp = int.to_bytes(int(obj.value), length=obj.nbytes, byteorder=endianess, signed=obj.signed)
        #     bsgy[obj.startbyte - 1:obj.startbyte + obj.nbytes - 1] = tmp
        # return bytes(bsgy)
        return self._to_bytes(endianess=endianess, byte_length=byte_length)


"""read and dump segy headers"""

import os
from pathlib import Path

import numpy
import pandas

from segytools.segy_file_header import SegyFileHeaderRev2
from segytools.segy_trace_header import SegyTraceHeaderRev2
from segytools.segpy.toolkit import unpack_ibm_floats, unpack_values

# modified from segpy toolkit.py
def read_binary_values(buf, num_items, ctype, endian):
    values = (unpack_ibm_floats(buf, num_items)
              if ctype == 'ibm'
              else unpack_values(buf, ctype, endian))
    assert len(values) == num_items
    return values


TEXTHEADERLENGTH = 3200
FILEHEADERLENGTH = 400
TRCHEADERLENGTH = 240
ENDIANESS = 'big'
textencode = 'ebcidic'

segyfile = "example.sgy"
pathsegyfile = Path(str(segyfile))
assert (pathsegyfile.is_file())

segyfilesize = os.path.getsize(segyfile)

bsgyout = bytes()
hdr4df = {}

# 'rb' is "read bytes"
with open(segyfile, 'rb') as fobj:
    # read the first 3200 bytes.
    # This will always be 3200 byte textual file header
    b_text_header = fobj.read(TEXTHEADERLENGTH)
    # print(type(b_text_header))

    textheader = str()
    if textencode == 'ebcidic':
        textheader = b_text_header.decode(encoding='cp500')
    elif textencode == 'ascii':
        textheader = b_text_header.decode(encoding='utf-8')

    # print(textheader)
    print('\nloading file header ...')
    fileheader = SegyFileHeaderRev2()
    b_file_header = fobj.read(FILEHEADERLENGTH)
    fileheader.set_file_header_values(bsgy=b_file_header, endianess=ENDIANESS)
    # print(fileheader)

    print('\ntrace data')
    #---TRCDATALENGTH = (fileheader.getKeyProperty('numsmp', 'value') * fileheader.getBytesPerSecond())
    TRCDATALENGTH = fileheader.number_of_samples() * fileheader.bytes_per_sample()
    print(f'trace data length in bytes: {TRCDATALENGTH}')

    TRCDATASTART = TEXTHEADERLENGTH + FILEHEADERLENGTH

    # assuming the trace length is the same for all traces. That is, number of samples and sample interval is consistent
    ntraces = (segyfilesize - (TEXTHEADERLENGTH + FILEHEADERLENGTH)) / (TRCDATALENGTH + TRCHEADERLENGTH)
    ntraces = int(ntraces)
    print('number of traces: ', ntraces)

    # loop through each trace (could be optimized later)
    for trccntr in range(ntraces):
        fobj.seek(TRCDATASTART + (trccntr * (TRCDATALENGTH + TRCHEADERLENGTH)))
        bsgy = fobj.read(TRCHEADERLENGTH)
        hdr = SegyTraceHeaderRev2()
        hdr.set_trace_header_values(bsgy=bsgy, endianess=ENDIANESS)
        # if trccntr == 0:
        #     print(hdr)
        hdr4df[str(trccntr)] = hdr.value_dict()

        # bsgy = fobj.read(TRCDATALENGTH)
        # trace_values = read_binary_values(bsgy, fileheader.numsmp.getValue(), fileheader.getCType(), endianess)
        # tracenp = numpy.asarray(trace_values, dtype=numpy.float32, order='C')
        # bsgyout += bsgy

    fobj.close()

# print('\ndataframe')
# print(hdr4df)

df = pandas.DataFrame(data=hdr4df)
df = df.transpose()
print(df.head())
print(df.tail())


'''
segy file header aka binary header
'''

# import python standard modules


# import 3rd party libraries


# import local python
from segytools.segy_header_item import SegyHeaderItem
from segytools.segy_abstract_header import SegyAbstractHeader
from segytools.datatypes import DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE, SEG_Y_TYPE_TO_CTYPE, size_in_bytes


# bytes 29-30
TRACE_SORTING_CODE = {
    -1: 'Other',
    0: 'Unknown',
    1: 'As Recorded (no sorting)',
    2: 'CDP Ensemble',
    3: 'Single fold continuous profile',
    4: 'Horizontally stacked',
    5: 'Common source point',
    6: 'Common receiver point',
    7: 'Common offset point',
    8: 'Common mid-point',
    9: 'Common Conversion point'
}

# bytes 39-40
SWEEP_TYPE_CODE = {
    1: 'linear',
    2: 'parabolic',
    3: 'exponential',
    4: 'other'
}

# bytes 47-48
TAPER_TYPE = {
    1: 'linear',
    2: 'cos**2',
    3: 'other'
}

# bytes 49-50
CORRELATED_DATA_TRACES = {
    1: 'no',
    2: 'yes'
}

# bytes 51-52
BINARY_GAIN_RECOVERED = {
    1: 'yes',
    2: 'no'
}

# bytes 53-54
AMPLITUDE_RECOVERY_METHOD = {
    1: 'none',
    2: 'spherical divergence',
    3: 'agc',
    4: 'other'
}

# bytes 55-56
MEASUREMENT_SYSTEM = {
    1: 'meters',
    2: 'feet'
}

# bytes 57-58
IMPULSE_SIGNAL_POLARITY = {
    1: 'incerase in pressure or upward geophone case movement gives negative number on tape',
    2: 'incerase in pressure or upward geophone case movement gives positive number on tape'
}

# bytes 59-60
VIBRATORY_POLARITY_CODE = {
    1: '337.5 (deg) to 22.5 (deg)',
    2: '22.5 (deg) to 67.5 (deg)',
    3: '67.5 (deg) to 112.5 (deg)',
    4: '112.5 (deg) to 157.5 (deg)',
    5: '157.5 (deg) to 202.5 (deg)',
    6: '202.5 (deg) to 247.5 (deg)',
    7: '247.5 (deg) to 292.5 (deg)',
    8: '292.5 (deg) to 337.5 (deg)'
}


class SegyFileHeaderRev2(SegyAbstractHeader):
    """
    File or binary header definition of a segy file.

    Parameters
    ----------
    jobid : SegyHeaderItem
        job identification number
    lineno : SegyHeaderItem
        line number
    reelno : SegyHeaderItem
        reel number
    ntrcens : SegyHeaderItem
        number of data traces per ensemble
    ntrcaux : SegyHeaderItem
        number of auxiliary traces per ensemble
    smpint : SegyHeaderItem
        sample interval in microseconds
    smpinto : SegyHeaderItem
        sample interval in microseconds or original recording
    numsmp : SegyHeaderItem
        number of samples per data trace
    numsmpo : SegyHeaderItem
        number of samples per data trace or original recording
    dsfmt : SegyHeaderItem
        data sample format code
    fold : SegyHeaderItem
        ensemble fold
    sortcode : SegyHeaderItem
        trace sorting code
    vsumcode : SegyHeaderItem
        vertical sum code
    sweepfs : SegyHeaderItem
        sweep frequency at start
    sweepfe : SegyHeaderItem
        sweep frequency at end
    sweeplen : SegyHeaderItem
        sweep length
    sweepcode : SegyHeaderItem
        sweep type code
    sweepchan : SegyHeaderItem
        trace_number of sweep channel
    sweeptprs : SegyHeaderItem
        sweep_trace taper length in ms at start
    sweeptpre : SegyHeaderItem
        sweep trace taper length in ms at end
    tprtype : SegyHeaderItem
        taper type
    corrtrc : SegyHeaderItem
        correlated data traces
    bingain : SegyHeaderItem
        binary gain recovered
    amprec : SegyHeaderItem
        amplitude recovery method
    meassys : SegyHeaderItem
        measurement system
    polarity : SegyHeaderItem
        impulse signal polarity
    vpolarity : SegyHeaderItem
        vibratory polarity code
    segyrev : SegyHeaderItem
        seg y format revision number
    fixedlen : SegyHeaderItem
        fixed length trace flag
    ntxthead : SegyHeaderItem
        number of 3200 byte ext file header records following
            

    Methods
    -------
    segy_type
        Returns the value of the data sample format. Result is one of `ibm`, `int32`, `int16`, `float32`, or `int8`. (see segpy datatypes)
    ctype
        Returns the format character used by the python standard library. (see segpy datatypes).
    bytes_per_sample
        Returns the number of bytes for a given datatype.
    number_of_samples
        Returns the number of samples from the `numsmp` header item.
    to_bytes
        Converts the container class to a bytes object usually for exporting a new segy file.
    """

    def __init__(self):
        super().__init__()
        self.jobid = SegyHeaderItem(
            name='jobid', 
            nbytes=4, 
            startbyte=1,
            description='job identification number',
            signed=True)
        self.lineno = SegyHeaderItem(
            name='lineno', 
            nbytes=4, 
            startbyte=5,
            description='line number',
            signed=True)
        self.reelno = SegyHeaderItem(
            name='reelno', 
            nbytes=4, 
            startbyte=9,
            description='reel number',
            signed=True)
        self.ntrcens = SegyHeaderItem(
            name='ntrcens', 
            nbytes=2, 
            startbyte=13,
            description='number of data traces per ensemble',
            signed=True)
        self.ntrcaux = SegyHeaderItem(
            name='ntrcaux', 
            nbytes=2, 
            startbyte=15,
            description='number of auxiliary traces per ensemble',
            signed=True)
        self.smpint = SegyHeaderItem(
            name='smpint', 
            nbytes=2, 
            startbyte=17,
            description='sample interval in microseconds',
            signed=True)
        self.smpinto = SegyHeaderItem(
            name='smpinto', 
            nbytes=2, 
            startbyte=19,
            description='sample interval in microseconds or original recording',
            signed=True)
        self.numsmp = SegyHeaderItem(
            name='numsmp', 
            nbytes=2, 
            startbyte=21,
            description='number of samples per data trace',
            signed=True)
        self.numsmpo = SegyHeaderItem(
            name='numsmpo', 
            nbytes=2, 
            startbyte=23,
            description='number of samples per data trace or original recording',
            signed=True)
        self.dsfmt = SegyHeaderItem(
            name='dsfmt', 
            nbytes=2, 
            startbyte=25,
            description='data sample format code',
            signed=True,
            map_bool=True, 
            map_dict=DATA_SAMPLE_FORMAT_TO_SEG_Y_TYPE)
        self.fold = SegyHeaderItem(
            name='fold', 
            nbytes=2, 
            startbyte=27,
            description='ensemble fold',
            signed=True)
        self.sortcode = SegyHeaderItem(
            name='sortcode', 
            nbytes=2, 
            startbyte=29,
            description='trace sorting code',
            signed=True,
            map_bool=True, 
            map_dict=TRACE_SORTING_CODE)
        self.vsumcode = SegyHeaderItem(
            name='vsumcode', 
            nbytes=2, 
            startbyte=31,
            description='vertical sum code',
            signed=True)
        self.sweepfs = SegyHeaderItem(
            name='sweepfs', 
            nbytes=2, 
            startbyte=33,
            description='sweep_frequency_at_start',
            signed=True)
        self.sweepfe = SegyHeaderItem(
            name='sweepfe', 
            nbytes=2, 
            startbyte=35,
            description='sweep_frequency_at_end',
            signed=True)
        self.sweeplen = SegyHeaderItem(
            name='sweeplen', 
            nbytes=2, 
            startbyte=37,
            description='sweep_length',
            signed=True)
        self.sweepcode = SegyHeaderItem(
            name='sweepcode', 
            nbytes=2, 
            startbyte=39,
            description='sweep_type_code',
            signed=True,
            map_bool=True, 
            map_dict=SWEEP_TYPE_CODE)
        self.sweepchan = SegyHeaderItem(
            name='sweepchan', 
            nbytes=2, 
            startbyte=41,
            description='trace_number_of_sweep_channel',
            signed=True)
        self.sweeptprs = SegyHeaderItem(
            name='sweeptprs', 
            nbytes=2, 
            startbyte=43,
            description='sweep_trace_taper_length_in_ms_at_start',
            signed=True)
        self.sweeptpre = SegyHeaderItem(
            name='sweeptpre', 
            nbytes=2, 
            startbyte=45,
            description='sweep_trace_taper_length_in_ms_at_end',
            signed=True)
        self.tprtype = SegyHeaderItem(
            name='tprtype', 
            nbytes=2, 
            startbyte=47,
            description='taper_type',
            signed=True,
            map_bool=True, 
            map_dict=TAPER_TYPE)
        self.corrtrc = SegyHeaderItem(
            name='corrtrc', 
            nbytes=2, 
            startbyte=49,
            description='correlated_data_traces',
            signed=True,
            map_bool=True, 
            map_dict=CORRELATED_DATA_TRACES)
        self.bingain = SegyHeaderItem(
            name='bingain', 
            nbytes=2, 
            startbyte=51,
            description='binary_gain_recovered',
            signed=True,
            map_bool=True, 
            map_dict=BINARY_GAIN_RECOVERED)
        self.amprec = SegyHeaderItem(
            name='amprec', 
            nbytes=2, 
            startbyte=53,
            description='amplitude_recovery_method',
            signed=True,
            map_bool=True, 
            map_dict=AMPLITUDE_RECOVERY_METHOD)
        self.meassys = SegyHeaderItem(
            name='meassys', 
            nbytes=2, 
            startbyte=55,
            description='measurement_system',
            signed=True,
            map_bool=True, 
            map_dict=MEASUREMENT_SYSTEM)
        self.polarity = SegyHeaderItem(
            name='polarity', 
            nbytes=2, 
            startbyte=57,
            description='impulse_signal_polarity',
            signed=True,
            map_bool=True, 
            map_dict=IMPULSE_SIGNAL_POLARITY)
        self.vpolarity = SegyHeaderItem(
            name='vpolarity', 
            nbytes=2, 
            startbyte=59,
            description='vibratory_polarity_code',
            signed=True,
            map_bool=True, 
            map_dict=VIBRATORY_POLARITY_CODE)
        #self.unassigned1 = SegyAbstractHeader(240, 'unassigned_1', False, 3261)
        self.segyrev = SegyHeaderItem(
            name='segyrev', 
            nbytes=2, 
            startbyte=301,
            description='seg_y_format_revision_number',
            signed=True)
        self.fixedlen = SegyHeaderItem(
            name='fixedlen', 
            nbytes=2, 
            startbyte=303,
            description='fixed_length_trace_flag',
            signed=True)
        self.ntxthead = SegyHeaderItem(
            name='ntxthead', 
            nbytes=2, 
            startbyte=305,
            description='number_of_3200_byte_ext_file_header_records_following',
            signed=True)
        #self.unassigned2 = SegyAbstractHeader(94, 'unassigned_2', False, 3507)


    def __str__(self):
        s = 'segy file headers\n'
        s += str(self.jobid) + '\n'
        s += str(self.lineno) + '\n'
        s += str(self.reelno) + '\n'
        s += str(self.ntrcens) + '\n'
        s += str(self.ntrcaux) + '\n'
        s += str(self.smpint) + '\n'
        s += str(self.smpinto) + '\n'
        s += str(self.numsmp) + '\n'
        s += str(self.numsmpo) + '\n'
        s += str(self.dsfmt) + '\n'
        s += str(self.fold) + '\n'
        s += str(self.sortcode) + '\n'
        s += str(self.vsumcode) + '\n'
        s += str(self.sweepfs) + '\n'
        s += str(self.sweepfe) + '\n'
        s += str(self.sweeplen) + '\n'
        s += str(self.sweepcode) + '\n'
        s += str(self.sweepchan) + '\n'
        s += str(self.sweeptprs) + '\n'
        s += str(self.sweeptpre) + '\n'
        s += str(self.tprtype) + '\n'
        s += str(self.corrtrc) + '\n'
        s += str(self.bingain) + '\n'
        s += str(self.amprec) + '\n'
        s += str(self.meassys) + '\n'
        s += str(self.polarity) + '\n'
        s += str(self.vpolarity) + '\n'
        s += str(self.segyrev) + '\n'
        s += str(self.fixedlen) + '\n'
        s += str(self.ntxthead)
        return s

    # --- METHODS ---
    def segy_type(self):
        """Return the segy type, e.g. 'ibm', 'int32', 'int16'.
        """
        return self.dsfmt.value

    def ctype(self):
        """Return the characther type used to convert the bytes to python data.
        """
        return SEG_Y_TYPE_TO_CTYPE[self.segy_type()]

    def bytes_per_sample(self):
        """Return the number of bytes per sample in a trace.
        """
        return size_in_bytes(self.ctype())

    def number_of_samples(self):
        """Return the number of samples in a trace.
        """
        return self.numsmp.value


    # MUTATORS
    def set_file_header_values(self, bsgy, endianess):
        """Set the file header items values by converting the bytes object into python data.
        """
        file_key_obj_dict = self.key_object_dict()
        for _, obj in file_key_obj_dict.items():
            bobj = bsgy[obj.startbyte - 1:obj.startbyte + obj.nbytes - 1]
            obj.value = int.from_bytes(bobj, byteorder=endianess, signed=obj.signed)

            # TODO: do not map by default. this will cause an error when converting back to bytes.
            if obj.map_bool is True:
                self.set_key_property(obj.name, 'value', self.mapped_value(obj.name))

    # def to_bytes(self, endianess, byte_length=400):
    #     bsgy = bytearray(byte_length)
    #     file_key_obj_dict = self.key_object_dict()
    #     for _, obj in file_key_obj_dict.items():
            
    #         if obj.map_bool is True:
    #             # reverse map back to integer
    #             # tmp_int = list(obj.map_dict.keys())[list(obj.map_dict.values()).index(obj.value)]
    #             tmp_int = 0
    #             for key, val in obj.map_dict.items():
    #                 if obj.value == val:
    #                     tmp_int = int(key)
    #                     break
    #             obj.value = tmp_int
    #         tmp = int.to_bytes(int(obj.value), length=obj.nbytes, byteorder=endianess, signed=obj.signed)
    #         bsgy[obj.startbyte - 1:obj.startbyte + obj.nbytes - 1] = tmp
    #     return bytes(bsgy)

    def to_bytes(self, endianess, byte_length=400):
        """Convert a SegyFileHeader object to bytes().
        """
        return self._to_bytes(endianess=endianess, byte_length=byte_length)

    # TODO: def is_fixed_length(self):
    #     if self.fixedlen == ??:
    #         return True
    #     else:
    #         return False
